import datetime
import os

from gwdg_idm_api.benutzerverwaltung import Benutzerverwaltung
from gwdg_idm_api.models import BaseGWDGUser, CreateTemplate
from gwdg_idm_api.util import AlreadyDeletedError, pretty_print


# This are the values for our CreateTemplate for a new IDM user
# Adjust the fields to your template.
class NewGWDGUser(CreateTemplate):
    create_template_name: str = os.environ["IDM_CREATE_TEMPLATE"]
    givenName: str
    sn: str
    workGroup: str
    departmentNumber: str
    roomNumber: str = ""
    telephoneNumber: str = ""
    goesternUserType: str = "0"
    externalTaskCommand: str = "CreateMailbox"


# This are the values for our MPIDoUser, we have some additional fields
# for workGroup, dn, and externalTaskCommand
# Adjust the fields to your template.
class MPIDoUser(BaseGWDGUser):
    workGroup: str
    dn: str
    externalTaskCommand: str


# Load idm credentials
# Do not upload credentials in any way to github/gitlab etc.
idm_username = os.environ["IDM_USERNAME"]
idm_password = os.environ["IDM_PASSWORD"]


# For testing purposes, we are using the idm-stage endpoint.
# This one is a copy of the previous day and can be used
# for sandbox testing.
# Change to "https://idm.gwdg.de/api/v3" for production.
benutzerverwaltung = Benutzerverwaltung(
    api_url=os.environ["IDM_API_URL"],
    username=idm_username,
    password=idm_password,
)
# Set the user class you would like to return
benutzerverwaltung.set_user_class(MPIDoUser)

# Get all users having their surename starting with 'a'
sn_a_users = benutzerverwaltung.get_multiple_users("$sn -eq 'a*'")
print("# First two entries starting with 'a'")
print(pretty_print(sn_a_users[:2]))

# Take the first user and print a more detailed view
print("# Get detailed information")
single_user = benutzerverwaltung.get_single_user(sn_a_users[0].id)
print(pretty_print(single_user))


# Change the lastname of the user to 'Upps bad lastnäme' and workGroup to '-'
# and check if the update happened
print("# Update user")
updated_user = benutzerverwaltung.update_user(
    single_user, {"workGroup": "-", "sn": "Upps bad lastnäme"}
)
print(pretty_print(updated_user))

# Ah that did not work out well!
# Now lets remove the user in 10 days (set goesternExpirationDate)
# If
print("# Delete user in 10 days")
try:
    deleted_user = benutzerverwaltung.delete_user(
        single_user, datetime.datetime.now() + datetime.timedelta(days=10)
    )
except AlreadyDeletedError:
    # If the user is already deleted an AlreadyDeletedError is thrown
    print("Already deleted")
else:
    print(pretty_print(deleted_user))

# Actually, lets delete him yesterday
# The delete_user functions will automatically just check the
# isScheduledForDeletion flag if the expire_datetime is in the past
print("# Delete user now")
try:
    deleted_user = benutzerverwaltung.delete_user(
        updated_user, datetime.datetime.now() - datetime.timedelta(days=1)
    )
except AlreadyDeletedError:
    print("Already deleted")
else:
    print(pretty_print(deleted_user))

# Lets create a new user
print("# Create new user")
new_user_tmp = NewGWDGUser(
    givenName="john", sn="doe", workGroup="-", departmentNumber="ZE"
)
new_user = benutzerverwaltung.create_user(new_user_tmp)
print(pretty_print(new_user))

print("# All done :)")



from gwdg_idm_api.util import username_prediction

predictions = [
    ("Markus", "Stabrin", "markus.stabrin"),  # Normal
    ("Markus", "Stabrinabcde", "m.stabrinabcde"),  # gt 18
    ("Markus", "Stabrinabcdefghijklmnop", "m.stabrinabcdefghi"),  # gt 18 after gt 18
    ("Markus", "Stabrinå", "markus.stabrina"),  # å to a
    ("Markus", "Stabrinß", "markus.stabrinss"),  # ß to ss
    ("Markus", "Stabrin\u0256", "markus.stabrind"),  # lowercase d bar to d
    ("Markus", "Stabrin s", "markus.stabrins"),  # space to no space
]

error = False
for given_name, surname, expected in predictions:
    prediction = username_prediction(given_name, surname)
    if prediction != expected:
        print(f"  Failed: {prediction=} != {expected=}")
        error = True
    else:
        print(f"Success: {given_name=} {surname=} {prediction=}")

if error:
    raise Exception("Errors occured")



import json
import typing

from . import string_cleaner

if typing.TYPE_CHECKING:
    from .models import BaseGWDGUser


class IDMNotReachableError(Exception):
    pass


class IDMRequestError(Exception):
    pass


class BadJsonError(Exception):
    pass


class AlreadyDeletedError(Exception):
    pass


class InvalidUIDException(Exception):
    pass


def pretty_print(obj: "BaseGWDGUser | list[BaseGWDGUser]") -> str:
    if isinstance(obj, list):
        return pretty_multy_print(obj)
    else:
        return pretty_single_print(obj)


def pretty_single_print(obj: "BaseGWDGUser") -> str:
    return json.dumps(obj.dict(), indent=2)


def pretty_multy_print(obj: list["BaseGWDGUser"]) -> str:
    return json.dumps([entry.dict() for entry in obj], indent=2)


def username_prediction(given_name, surname):
    given_name = string_cleaner.clean(given_name)
    surname = string_cleaner.clean(surname)
    if not given_name and not surname:
        raise InvalidUIDException(
            "Givenname oder Surname ist nach dem StringClean leer und es kann daher kein username aus givenname.surname generiert werden."
        )

    username = f"{given_name}.{surname}"
    if len(username) > 18:
        username = f"{given_name[0]}.{surname}"

    if len(username) > 18:
        username = f"{given_name[0]}.{surname[:16]}"

    return username



import datetime
import json
import typing
import urllib.parse

from .models import BaseGWDGUser, ChangeTemplate, CreateTemplate, IDMRequest
from .util import AlreadyDeletedError, BadJsonError, IDMRequestError

if typing.TYPE_CHECKING:
    from requests import Response


class Benutzerverwaltung(IDMRequest):
    api_suffix: str = "/Benutzerverwaltung/objects"
    user_class: type[ChangeTemplate] = BaseGWDGUser

    def set_user_class(self, user_class: type[ChangeTemplate]):
        self.user_class = user_class

    def get_single_user(self, user_id: str) -> ChangeTemplate:
        request_url: str
        resp: "Response"
        json_resp: dict

        request_url = f"{self.api_url}{self.api_suffix}/{user_id}"
        resp = self.get_request(request_url)
        try:
            json_resp = json.loads(resp.text)
        except json.decoder.JSONDecodeError:
            raise BadJsonError(resp.text)
        return self.user_class.from_json(json_resp)

    def get_multiple_users(
        self, filter_string: str | None = None
    ) -> list[ChangeTemplate]:
        request_url: str
        resp: "Response"
        json_resp: dict

        request_url = f"{self.api_url}{self.api_suffix}"
        if filter_string is not None:
            request_url = f"{request_url}?filter={urllib.parse.quote(filter_string)}"
        resp = self.get_request(request_url)
        try:
            json_resp = json.loads(resp.text)
        except json.decoder.JSONDecodeError:
            raise BadJsonError(resp.text)
        return [self.user_class.from_json(obj) for obj in json_resp["Objects"]]

    def update_user(
        self, user: ChangeTemplate, update_dict: dict[str, str | list[str]]
    ) -> ChangeTemplate:
        request_url: str = f"{self.api_url}{self.api_suffix}/{user.id}"
        resp: "Response"
        data: str

        if "mail" in update_dict:
            # Run mail at the very last to avoid missing proxyMail entries
            mail = update_dict.pop("mail")
            update_dict["mail"] = mail

        for key, value in update_dict.items():
            data = user.update_json(key, value)

            resp = self.put_request(request_url, data)
            if "success" not in resp.text:
                raise IDMRequestError(
                    f"Could not connect to IDM: Invalid combination\n{key=}\n{value=}\n{resp.text=}"
                )
        return self.get_single_user(user.id)

    def create_user(self, new_user: CreateTemplate) -> ChangeTemplate:
        request_url: str = (
            f"{self.api_url}{self.api_suffix}/{new_user.create_template_name}"
        )
        resp: "Response"
        data: str

        data = new_user.to_json()
        resp = self.post_request(request_url, data)
        if "success" not in resp.text:
            raise IDMRequestError(
                f"Could not connect to IDM: Invalid combination\n{resp.text=}"
            )
        user_id = resp.headers["location"]
        return self.get_single_user(user_id)

    def delete_user(self, user: ChangeTemplate, expire_datetime: datetime.datetime):
        today: datetime.date
        expire_date: datetime.date
        activate_now: bool
        change_data: dict[str, str]

        current_user_status = self.get_single_user(user.id)
        if current_user_status.goesternUserStatus in ("2", "255"):
            raise AlreadyDeletedError

        change_data = {}
        if current_user_status.isScheduledForDeletion == "TRUE":
            change_data["isScheduledForDeletion"] = "FALSE"
        elif current_user_status.goesternExpirationDate:
            change_data["goesternExpirationDate"] = ""

        self.update_user(user, change_data)  # type: ignore

        today = datetime.datetime.now().date()
        expire_date = expire_datetime.date()

        activate_now = (expire_date - today).total_seconds() <= 0
        if activate_now:
            change_data = {
                "isScheduledForDeletion": "TRUE",
            }
        else:
            change_data = {
                "goesternExpirationDate": expire_date.strftime("%d.%m.%Y"),
            }

        return self.update_user(user, change_data)  # type: ignore



import json
import logging
import os
import typing

import pydantic
import requests

from .util import IDMNotReachableError

if typing.TYPE_CHECKING:
    from requests import Response


__all__ = ["BaseGWDGUser"]


DATA_TYPE = str | list | dict | None


class IDMRequest(pydantic.BaseModel):
    username: str
    password: str
    api_url: str = os.environ.get("IDM_API_URL", "")

    def run_request(
        self, func: typing.Callable[..., "Response"], url: str, data: DATA_TYPE = None
    ) -> "Response":
        try:
            return func(
                url,
                auth=(self.username, self.password),
                timeout=60,  # Needs to be this high to allow for an all-objects query
                headers={
                    "Accept": "application/json",
                    "Accept-Language": "en",
                    "Content-Type": "application/json",
                },
                data=data,
            )
        except Exception as e:
            raise IDMNotReachableError(
                f"Could not connect to IDM: IDM not reachable!\n{e}"
            )

    def get_request(self, url: str) -> "Response":
        return self.run_request(requests.get, url)

    def put_request(self, url: str, data: DATA_TYPE) -> "Response":
        return self.run_request(requests.put, url, data)

    def post_request(self, url: str, data: DATA_TYPE) -> "Response":
        return self.run_request(requests.post, url, data)


class ChangeTemplate(pydantic.BaseModel):
    id: str
    goesternExpirationDate: str = None
    isScheduledForDeletion: str = None
    goesternUserStatus: str = None

    @staticmethod
    def update_json(name: str, value: str | list[str]) -> str:
        data = {
            "name": name,
            "value": [value] if not isinstance(value, list) else value,
        }
        return json.dumps({"attributes": [data]})

    @classmethod
    def from_json(cls, json: dict) -> "ChangeTemplate":
        response_dict: dict[str, list[str] | str]

        response_dict = {
            "id": [json["id"]],
            "dn": [json["dn"]],
        }
        response_dict.update(
            {entry["name"]: entry["value"] for entry in json["attributes"]}
        )

        remove_keys = []
        for key, value in response_dict.items():
            try:
                expected_type = type(cls.__fields__[key].outer_type_())
            except KeyError:
                remove_keys.append(key)
                logging.debug(
                    "\n"
                    "  key unknown to model\n"
                    f"  User: {response_dict['id']}\n"
                    f"  Key: {key}"
                    f"  Value: {value}"
                )
                continue
            if isinstance(value, expected_type):
                continue

            if isinstance(value, expected_type):
                continue
            elif expected_type is str and isinstance(value, list):
                if len(value) > 1:
                    logging.warning(
                        "\n"
                        "  str expected, but found list: Using first element\n"
                        "  Please check your class specifications.\n"
                        f"  User: {response_dict['id']}\n"
                        f"  Key: {key}"
                        f"  Value: {value}"
                    )
                try:
                    new_val: str = value[0]
                except IndexError:
                    logging.warning(
                        "  str expected, but empty list found: Set to empty string\n"
                        "  Please check your class specifications.\n"
                        f"  User: {response_dict['id']}\n"
                        f"  Key: {key}"
                        f"  Value: {value}"
                    )
                    new_val: str = ""
                response_dict[key] = new_val
            else:
                assert False, (
                    "  Only str and list types are supported so far!"
                    "  Please check your class specifications.\n"
                    f"  User: {response_dict['id']}\n"
                    f"  Key: {key}"
                    f"  Value: {value}"
                )

        return cls(**{key: value for key, value in response_dict.items() if key not in set(remove_keys)})  # type: ignore


class CreateTemplate(pydantic.BaseModel):
    create_template_name: str

    def to_json(self) -> str:
        data = [
            {
                "name": key.removeprefix("_"),
                "value": [value] if not isinstance(value, list) else value,
            }
            for key, value in self.dict().items()
            if key != "create_template_name"
        ]
        return json.dumps({"attributes": data})


class BaseGWDGUser(ChangeTemplate):
    # Common fields
    ou: str = None
    employeeNumber: str = None
    mpgEmployeeNumber: str = None
    employeeType: str = None
    employeeStatus: str = None
    accountType: str = None
    uid: str = None
    oldUid: list[str] = None
    goesternSAMAccountName: str = None
    ou: str = None
    goesternUserType: str = None
    givenName: str = None
    sn: str = None
    goesternGWDGadDisplayName: str = None
    description: str = None
    departmentNumber: str = None
    title: str = None
    telephoneNumber: str = None
    mobile: str = None
    facsimileTelephoneNumber: str = None
    roomNumber: str = None
    street: str = None
    postalCode: str = None
    city: str = None
    st: str = None
    l: str = None
    goesternProxyAddresses: list[str] = None
    mail: str = None
    goesternExchangeQuota: str = None
    goesternMailboxServer: str = None
    goesternMailboxZugehoerigkeit: str = None
    exchangeTargetAddress: str = None
    goesternMailRoutingAddresses: list[str] = None
    goesternExchHideFromAddressLists: str = None
    externalEmailAddress: list[str] = None
    filterAttribute1: list[str] = None
    filterAttribute2: list[str] = None
    filterAttribute3: list[str] = None
    goesternDisableReason: str = None
    goesternDisableDate: str = None
    goesternRemoveDate: str = None
    goesternLockoutTime: str = None
    ownCloudQuota: str = None
    memberOfStaticExchangeDistGrp: list[str] = None
    isInitialPassword: str = None
    createTimestamp: str = None
    modifyTimeStamp: str = None
    pwdChangedTime: str = None
    passwordExpirationTime: str = None
    isInitialAdditionalPassword: str = None
    additionalPasswordModifyTime: str = None
    additionalPasswordExpirationTime: str = None
    eduPersonPrincipalName: str = None
    effectivePrivilege: list[str] = None
    responsiblePerson: list[str] = None
